// версия 0 - время из NTP, индикатор MT1637 (статья из инта)
// версия 1 от 06.09.2019г. - индикатор заменен на матричный на MAX7219 - 4 шт., добавлена DS3231 для счета времени (базовая версия)
// версия 2 от 22.12.2019г. - индикатор на MAX7219 расширен до 8 разрядов (обр. совместимо)
// версия 3 от 05.10.2022г. - введен код для подключения большого семисегментного индикатора управляемый сдвиговыми регистрами (обр. совместимо)
// версия 4 от 06.10.2022г. - добавлена возможноть подключения датчиков температуры DS18B20 
// версия 5 от 19.10.2022г. - добавлен веб-сервер и настройки через веб-интерфейс 

#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <ESP8266WebServer.h>
#include <Wire.h>
#include <OneWire.h>
#include <EEPROM.h>
#include "Time.h"
#include "LedControl.h"
#define ESP8266

// НАСТРОЙКИ ПО УМОЛЧАНИЮ
/*#define AP_SSID         "TEST"              // SSID Wi-Fi сети
#define AP_PASS           "00000000000"       // пароль на Wi-Fi сеть
#define TIMEZONE          7         // Временная зона
#define SYNC_TIME_PERIOD  60        // Период синхронизации времени, минут
#define IND_BRIGHT        5         // яркость семисегментного дисплея в % от макс, значения от 1 до 100
#define MAX7219_BRIGHT    1         // яркость матричного дисплея, значения от 0 до 15
#define DS18B20_MAX_COUNT 2         // максимальное количество датчиков ds18b20 на шине
#define TEMPER_IND_PERIOD 20        // продолжительность всего цикла индикации, сек
#define TEMPER_IND_DURAT  2         // длительность отображения каждой температуры, сек 
*/
const char*   ssid =      "ESP-CLOCK";  // собственной сети
const char*   password =  "0123456789"; // собственной сети
#define DS18B20_MAX_COUNT 2         // максимальное количество датчиков ds18b20 на шине

// КУДА, ЧТО, КАК ПОДКЛЮЧЕНО
// пины WEMOS D1 mini (маркировка на плате Dx - связь с пинами GPIOx)
#define D0                16  //WAKE, ШИМ невозможен
#define D1                5
#define D2                4
#define D3                0   //FLASH
#define D4                2   // TXD1 - ВНИМАНИЕ! использовать только как ВЫХОД! иначе ЕСП не программируется, виснет и на GPIO0 (D3) появляется синус 26МГц
#define D5                14
#define D6                12
#define D7                13  //RXD2
#define D8                15  //TXD2
#define D9                3   //RXD0  - занят УАРТом
#define D10               1   //TXD0  - занят УАРТом
// подключение матричного дисплея
#define MAX7219_DIN       D2
#define MAX7219_CLK       D4
#define MAX7219_CS        D3
#define MAX7219_COUNT     8    // количество модулей матричного дисплея на MAX7219
// подключение микросхемы часов реального времени DS3231
#define DS3231_SCL        D5
#define DS3231_SDA        D6
#define DS3231_I2C_ADR    0x68 // адрес микросхемы на и2ц шине
// подключение семисегментного дисплея
#define IND_D             D0
#define IND_C1            D8
#define IND_C2            D7
// подключение DS18B20
#define DS18B20_D         D1

// представление цифр 0 до 10 для Dot matrix module MAX7219 в виде двоичных массивов
byte digit[10][8]=
{B00001110, B00010001, B00010001, B00010001, B00010001, B00010001, B00010001, B00001110,
 B00000010, B00000110, B00001010, B00000010, B00000010, B00000010, B00000010, B00001111,
 B00001110, B00010001, B00000001, B00000010, B00000100, B00001000, B00010000, B00011111,
 B00001110, B00010001, B00000001, B00000110, B00000001, B00000001, B00010001, B00001110,
 B00000001, B00000011, B00000101, B00001001, B00010001, B00011111, B00000001, B00000001,
 B00011110, B00010000, B00010000, B00001110, B00000001, B00000001, B00010001, B00001110,
 B00001110, B00010000, B00010000, B00011110, B00010001, B00010001, B00010001, B00001110,
 B00011111, B00000001, B00000001, B00000010, B00000100, B00001000, B00001000, B00001000,
 B00001110, B00010001, B00010001, B00001110, B00010001, B00010001, B00010001, B00001110,
 B00001110, B00010001, B00010001, B00010001, B00001111, B00000001, B00000001, B00001110};
// представление двух точек  для вывода  формата ЧЧ:MM 
byte dot[8]={B00000000, B11000000, B11000000, B00000000, B00000000, B11000000, B11000000, B00000000};
// представление точки  для вывода десятичного дробного формата 
byte pset[8]={B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B11000000, B11000000};
// представление прочерков
byte minus[8]={B00000000, B00000000, B00000000, B00011111, B00011111, B00000000, B00000000, B00000000};
// представление градуса
byte gradus[8]={B00001110, B00010001, B00010001, B00010001, B00001110, B00000000, B00000000, B00000000};
// представление букв дней недели (по 2 буквы)
byte day_week1[7][8]=
{ 0b00011110, 0b00010001, 0b00010001, 0b00011110, 0b00010001, 0b00010001, 0b00010001, 0b00011110, 
  0b00001111, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 
  0b00011110, 0b00010001, 0b00010001, 0b00011110, 0b00010001, 0b00010001, 0b00010001, 0b00011110, 
  0b00001110, 0b00010001, 0b00010000, 0b00010000, 0b00010000, 0b00010000, 0b00010001, 0b00001110, 
  0b00010001, 0b00010001, 0b00010001, 0b00001111, 0b00000001, 0b00000001, 0b00000001, 0b00000001, 
  0b00001111, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 0b00001001, 
  0b00001110, 0b00010001, 0b00010000, 0b00010000, 0b00010000, 0b00010000, 0b00010001, 0b00001110};
byte day_week2[7][8]=
{ 0b00000000, 0b00000000, 0b00000000, 0b00110000, 0b01000000, 0b01000000, 0b01000000, 0b00110000, 
  0b00000000, 0b00000000, 0b00000000, 0b00100100, 0b00100100, 0b00111100, 0b00100100, 0b00100100, 
  0b00000000, 0b00000000, 0b00000000, 0b01111100, 0b00010000, 0b00010000, 0b00010000, 0b00010000, 
  0b00000000, 0b00000000, 0b00000000, 0b00111000, 0b00100100, 0b00111000, 0b00100000, 0b00100000, 
  0b00000000, 0b00000000, 0b00000000, 0b01111100, 0b00010000, 0b00010000, 0b00010000, 0b00010000, 
  0b00000000, 0b00000000, 0b00000000, 0b01111100, 0b00010000, 0b00010000, 0b00010000, 0b00010000, 
  0b00000000, 0b00000000, 0b00000000, 0b01110000, 0b01000000, 0b01110000, 0b01001000, 0b01110000};

// представление цифр 0 до F для семисегментного индикатора
const byte ind_code[16] = {0xE7, 0x81, 0x73, 0xD3, 0x95, 0xD6, 0xF6, 0x83, 0xF7, 0xD7, 0xB7, 0xF4, 0x66, 0xF1, 0x76, 0x36};
/*
      размещение сегментов на индикаторе
                     A
                   #####
                  #     #
                F-#     #-B
                  #     #
                 G-#####
                  #     #
                E-#     #-C
                  #     #
                   #####   ##
                    D      #-H
*/

LedControl  my_MAX7219 =     LedControl(MAX7219_DIN, MAX7219_CLK, MAX7219_CS, MAX7219_COUNT); //инициализация дисплеев (MAX7219_COUNT шт.)

OneWire  ds18b20(DS18B20_D);

byte ds3231_damp[19];
byte ds18b20_addr[DS18B20_MAX_COUNT][8];
int ds18b20_Tx10[DS18B20_MAX_COUNT];
byte ds18b20_count = 0;

struct 
{
  char ap_ssid[20];          // SSID Wi-Fi сети
  char ap_pass[20];          // пароль на Wi-Fi сеть
  char ntp[30];              // НТП-сервер
  
  byte time_zone;            // Временная зона
  byte sync_time_period;     // Период синхронизации времени, часов
  byte seg_ind_bright;       // яркость семисегментного дисплея в % от макс, значения от 1 до 100
  byte matrix_ind_bright;    // яркость матричного дисплея, значения от 0 до 15
  byte ind_period;           // продолжительность всего цикла индикации, сек
  byte temper_int_duration;  // длительность отображения каждой температуры, сек
} setting;

struct 
{
  byte second;     // сек
  byte minutes;    // мин
  byte hours;      // час
  byte day;      // день (1..7)
  byte date;     // число (1..31)
  byte month;      // месяц (1..12)
  byte year;     // год (0..99)

  signed char temper;    // температура
} ds3231_time, autonom_time, now_time;

unsigned int  localPort = 2390;      // local port to listen for UDP packets
unsigned long ntp_time = 0;           // для хранения времени от НТП сервера
unsigned long cur_ms   = 0;           // время с момента включения, мс
unsigned long ms1      = millis();    // время включения
unsigned long t_cur    = 0;           // время с момента включения, сек
bool          points   = true;        // флаг включения двоеточия-разделителя
unsigned int err_count = 0;           // счетчик ошибок

unsigned long ms_syncro = 0;          // хранит время с момента включения в момент синхронизации с НТП, мс

const int NTP_PACKET_SIZE = 48; 
byte packetBuffer[ NTP_PACKET_SIZE];

IPAddress timeServerIP; 
ESP8266WebServer web_server(80);
WiFiUDP udp;


// *******************************************************************************************************************************
// СЛУЖЕБНЫЕ ФУНКЦИИ 
// *******************************************************************************************************************************

byte bcd_to_bin(byte inn) {  return ( (inn & 0x0F) + 10*( (inn>>4) & 0x0F) ); }

byte bin_to_bcd(byte inn) {  return ((inn / 10) << 4) | (inn % 10); }

// вычисляет текущее время
void calc_time()
{
// автономное вычисление текущего времени, после того как однажды засинхронизировались    
  unsigned long now_time_ = ntp_time + (millis() - ms_syncro)/1000;  // вычисляем текущее unix-время
  autonom_time.second = second (now_time_);        // the second now (0-59)
  autonom_time.minutes = minute(now_time_);        // the minute now (0-59)
  autonom_time.hours = hour(now_time_);            // the hour now  (0-23)
  autonom_time.date = day(now_time_);              // the day now (1-31)
  autonom_time.day = weekday(now_time_);           // day of the week (1-7), Sunday is day 1
  autonom_time.month = month(now_time_);           // the month now (1-12)
  autonom_time.year = year(now_time_)-2000;        // the full four digit year: (2009, 2010 etc)

// читаем м/с ds3231, если прочиталась, то ее время присваиваем в текущее, иначе за текущее берем автономное
  if ( ds3231_read(false) ) now_time = ds3231_time;
  else now_time = autonom_time;
}

// *******************************************************************************************************************************
// РАБОТА С СЕТЬЮ 
// *******************************************************************************************************************************

// Подключение к вафле
bool ConnectWiFi(const char *ssid, const char *pass) 
{
// Три попытки соединения по WiFi
   for( int i=0; i<3; i++){
      Serial.print("\nConnecting to: "); 
      Serial.println(ssid); 
      WiFi.begin(ssid,pass);
      delay(1000);
// Максиммум 12 раз проверка соединения       
      for( int j=0; j<12; j++ ){
          if (WiFi.status() == WL_CONNECTED) { 
              Serial.print("\nWiFi connect true: ");
              Serial.print(WiFi.localIP());
              Serial.print("/");
              Serial.print(WiFi.subnetMask());
              Serial.print("/");
              Serial.println(WiFi.gatewayIP());
              return true; 
          } 
          delay(1000);
          Serial.print(WiFi.status()); 
      }
   }   
   Serial.println("\nConnect WiFi failed ...");
   return false;
} 

// Запрос времени
bool GetNTP(const char *ntp_serv) 
{
// Отправка DNS запроса
  WiFi.hostByName(ntp_serv, timeServerIP); // 
//  Serial.print("Sending NTP packet ...");
// Очистка буфера в 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
// Формируем строку зыпроса NTP сервера
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;
// Посылаем запрос на NTP сервер (123 порт)
  udp.beginPacket(timeServerIP, 123); 
  udp.write(packetBuffer, NTP_PACKET_SIZE);
  udp.endPacket();

  delay(100);

// проверка наличия ответа, получение ответа
  int cb = udp.parsePacket();
  if (!cb) 
  {
    Serial.print(".");
    return false;
  }
  else 
  {
//    Serial.print(" packet length=");
//    Serial.print(cb);
// Читаем пакет в буфер    
    udp.read(packetBuffer, NTP_PACKET_SIZE); 
// 4 байта начиная с 40-го сождержат таймстамп времени - число секунд от 01.01.1900   
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
// Конвертируем два слова в переменную long
    unsigned long secsSince1900 = highWord << 16 | lowWord;
// Конвертируем в UNIX-таймстамп (число секунд от 01.01.1970
    const unsigned long seventyYears = 2208988800UL;
    unsigned long epoch = secsSince1900 - seventyYears;
// Делаем поправку на местную тайм-зону
    ntp_time = epoch + setting.time_zone*3600;    
    
//  запоминаем момент синхронизации c НТП по внутренней шкале (нужно для вычисления автономного времени)
    ms_syncro = millis(); 

  }
  return true;
}

// *******************************************************************************************************************************
// РАБОТА С DS3231 (М/С ЧАСОВ РЕАЛЬНОГО ВРЕМЕНИ)
// *******************************************************************************************************************************

void ds3231_set_of_NTP()
{
  ds3231_time.second = second(ntp_time);         // the second now (0-59)
  ds3231_time.minutes = minute(ntp_time);        // the minute now (0-59)
  ds3231_time.hours = hour(ntp_time);            // the hour now  (0-23)
  ds3231_time.date = day(ntp_time);              // the day now (1-31)
  ds3231_time.day = weekday(ntp_time);           // day of the week (1-7), Sunday is day 1
  ds3231_time.month = month(ntp_time);           // the month now (1-12)
  ds3231_time.year = year(ntp_time)-2000;        // the full four digit year: (2009, 2010 etc)

  ds3231_write();
}

bool ds3231_read(bool debug_info)
{
  byte i = 0;
  byte err = 0xFF; 
  
  Wire.beginTransmission(DS3231_I2C_ADR); // передача на устройство с адресом DS3231_I2C_ADR 
  Wire.write(0x00);                       // передаём начальный адрес чтения
  Wire.endTransmission(true);             // передача

  Wire.requestFrom(DS3231_I2C_ADR, sizeof(ds3231_damp)); // запрос на чтение
  while ( i < sizeof(ds3231_damp) ) 
  {
    ds3231_damp[i] = Wire.read(); // читаем все в массив, затем его разбираем
    err = err & ds3231_damp[i];
    i++;
  }

  if (err == 0xFF) 
  {
    ds3231_time.second = 0;
    ds3231_time.minutes = 0;
    ds3231_time.hours = 0;
    ds3231_time.day = 0 ;
    ds3231_time.date = 0;
    ds3231_time.month = 0;
    ds3231_time.year = 0;
    ds3231_time.temper = -100;
    if (debug_info) Serial.println("ds3231_read: Error: RTC DS3231 not connected ");
    return false;
  }

  ds3231_time.second = bcd_to_bin( ds3231_damp[0x00] );
  ds3231_time.minutes = bcd_to_bin( ds3231_damp[0x01] );
  ds3231_time.hours = bcd_to_bin( ds3231_damp[0x02] );
  ds3231_time.day = ds3231_damp[0x03] ;
  ds3231_time.date = bcd_to_bin( ds3231_damp[0x04] );
  ds3231_time.month = bcd_to_bin( ds3231_damp[0x05] );
  ds3231_time.year = bcd_to_bin( ds3231_damp[0x06] );
  ds3231_time.temper =ds3231_damp[0x11];

  return true;
}

void ds3231_write()
{
  ds3231_damp[0x00] = bin_to_bcd( ds3231_time.second );
  ds3231_damp[0x01] = bin_to_bcd( ds3231_time.minutes );
  ds3231_damp[0x02] = bin_to_bcd( ds3231_time.hours ) & 0b00111111; // 24 часовой формат
  ds3231_damp[0x03] = ds3231_time.day;
  ds3231_damp[0x04] = bin_to_bcd( ds3231_time.date );
  ds3231_damp[0x05] = bin_to_bcd( ds3231_time.month );
  ds3231_damp[0x06] = bin_to_bcd( ds3231_time.year );
  ds3231_damp[0x0E] = 0b00000000; // 1 Гц, без прерываний, выход активен только при основном питании
  ds3231_damp[0x0F] = 0b00001000; // 32.768 кГц включено

  Wire.beginTransmission(DS3231_I2C_ADR);         // передача на устройство с адресом DS3231_I2C_ADR 
  Wire.write(0x00);                               // начальный адрес
  Wire.write(ds3231_damp, sizeof(ds3231_damp)-3);   // весь массив, кроме 3 последних байт: подстройка частоты и температуры
  Wire.endTransmission();                         // старт передачи
}


// *******************************************************************************************************************************
// РАБОТА С DS18Х20 (ДАТЧИКИ ТЕМПЕРАТУРЫ)
// *******************************************************************************************************************************

// инициализация шины и поиск устройств на шине, считанные адреса записываются в ds18b20_addr[]
void search_ds18b20()   
{
 byte n = 0;
 byte temp_rom[8];
 
 Serial.println("\n\nInitializing module ds18b20");
 Serial.print("Max count device = "); Serial.println(DS18B20_MAX_COUNT);
 Serial.println("Found devices:");
 
 ds18b20_count = 0;
 for (byte i=0; i<DS18B20_MAX_COUNT; i++) ds18b20_Tx10[i] = -1000;  // заполняем массив кодом ошибки (запредельным значением)
 
 while ( ds18b20.search(temp_rom)  )              // пока что-то находится
 {
   if (OneWire::crc8(temp_rom, 7) != temp_rom[7]) { Serial.println("search_ds18b20(): CRC error!"); continue; } // проверяем соответствие црц
   
   Serial.print(n+1); Serial.print(") ROM = ");   // выводим адрес
   for (byte i=0; i<8; i++) 
   {
      if (temp_rom[7-i]<0x10) Serial.print(0);    // если один знак, то выводим доп ноль спереди
      Serial.print(temp_rom[7-i], HEX); 
   }
   
   switch (temp_rom[0])                           // выводим тип датчика
   {  
     case 0x10: Serial.println(" / Chip DS18S20 or old DS1820"); break;
     case 0x28: Serial.println(" / Chip DS18B20"); break;
     case 0x22: Serial.println(" / Chip DS1822"); break;
     default: Serial.println(" No device of DS18x20 type"); break;
   }
   if (n<DS18B20_MAX_COUNT) for (byte i=0; i<8; i++) ds18b20_addr[n][i] = temp_rom[i];
   n++;
 }

 Serial.println();
 ds18b20.reset_search();
 ds18b20_count = n;
 if (n>DS18B20_MAX_COUNT) ds18b20_count = DS18B20_MAX_COUNT;  // если датчиков больще чем ограничение, то ограничиваем

 ds18b20.reset();
 ds18b20.write(0xCC); // пропук адресации - обращаемся ко всем датчикам
 ds18b20.write(0x44); // команда старт измерения температуры
 
 return; 
}

// считывает значения температур (умноженных на 10) в массив ds18b20_Tx10
void read_T_ds18b20() 
{
  byte data[12];

  for (byte n=0; n<ds18b20_count; n++)
  {
    ds18b20.reset();
    ds18b20.select(ds18b20_addr[n]);   // Адресуем n-й датчик
    ds18b20.write(0xBE);          // Читаем память

    ds18b20_Tx10[n] = -1000;
    
    for (byte i=0; i<9; i++) data[i] = ds18b20.read();

    if ( OneWire::crc8(data, 8) != data[8] ) { Serial.println("read_T_ds18b20(): CRC error!"); continue;}

    int16_t raw = (data[1] << 8) | data[0];
    
    // датчик может быть настроен на разную точность, выясняем её 
    byte cfg = (data[4] & 0x60);
    if (cfg == 0x00) raw = raw & ~7; // точность 9-разрядов, 93,75 мс
    else if (cfg == 0x20) raw = raw & ~3; // точность 10-разрядов, 187,5 мс
    else if (cfg == 0x40) raw = raw & ~1; // точность 11-разрядов, 375 мс

    ds18b20_Tx10[n] = (float)raw / 1.6; // преобразование показаний датчика в градусы Цельсия 
  }
  
  ds18b20.reset();
  ds18b20.write(0xCC); // пропук адресации - обращаемся ко всем датчикам
  ds18b20.write(0x44); // команда старт измерения температуры
}




// *******************************************************************************************************************************
// РАБОТА С МАТРИЧНЫМ ИНДИКАТОРОМ
// *******************************************************************************************************************************

// выводит на  LED Dot matrix module MAX7219  цифру "dig"  в разряд "col"
void Write_One_Digit (int dig, int col)
{
byte psett = 0;

if (dig & 0x80)   // если старший разряд взведен - точка перед цифрой
{
  dig = dig & 0x7F;           // обнуляем бит точки
  psett = 1;                  // флаг точи
}

for (int t=0; t<8; t++)
   {
      byte a = 0, b = 0;
      
      if (dig == '*'){ my_MAX7219.setRow(col,t, gradus[t]); continue; } // если знак градуса
      if (dig > 9){ my_MAX7219.setRow(col,t, minus[t]); continue; }     // если не знак градуса и не цифра - то прочерк
      if ( (col == 3) && (dig == 0) ) { my_MAX7219.setRow(col,t, 0); continue; } // гащение незначащего нуля в 3м разряде
        
      if ( ((col == 1) || (col == 5)) && points ) a = dot[t]; // если разряд 1 и включено двоеточие то подмешиваем код двоеточия 
      if (psett) a = pset[t]; // если в разряде точка то подмешиваем код точки 
      if ((col == 2) || (col == 3) || (col == 6) || (col == 7) ) b = 1;    // если разряд 2, 3, 6, 7 и больше то сдвигаем цифры на 1 позицию влево
      my_MAX7219.setRow(col,t, digit[dig][t]<<b | a);
   }
}

// выводит на LED Dot matrix module MAX7219 две буквы дня недели в разряд "col" и "col+1"
void Write_Day (int day, int col)
{ 
for (int t=0; t<8; t++)
   {
      my_MAX7219.setRow(col,t, day_week2[day-1][t]);
      my_MAX7219.setRow(col+1,t, day_week1[day-1][t]);
   }
}


// Индикация времени на матричном дисплее
void print_time()
{
  Write_One_Digit (now_time.hours/10, 3);
  Write_One_Digit (now_time.hours%10, 2);
  Write_One_Digit (now_time.minutes/10, 1);
  Write_One_Digit (now_time.minutes%10, 0);

  Write_One_Digit (now_time.date/10, 7);
  Write_One_Digit (now_time.date%10, 6);

  Write_Day(now_time.day, 4);  
} 

// Индикация температуры (и времени) на матричном дисплее
void print_termo(int Tx10)
{
  byte a[4];
  const byte PSET = 0x80;

  // включаем двоеточие и выводим время туда где была дата и день недели (если дисплей всего 4х разрядный, то в небытие)
  points = 1;                             
  Write_One_Digit (now_time.hours/10, 7);
  Write_One_Digit (now_time.hours%10, 6);
  Write_One_Digit (now_time.minutes/10, 5);
  Write_One_Digit (now_time.minutes%10, 4);

  points = 0;                             // отключаем двоеточие

  if ((Tx10 > 1250) || (Tx10 < -550))               // если выход за диапазон измерений
  {
    Write_One_Digit ('-', 0);
    Write_One_Digit ('-', 1);
    Write_One_Digit ('-', 2);
    Write_One_Digit ('-', 3);
    return;
  }

  a[0] = '*';
  
  if ( (Tx10 >= 0) && (Tx10 < 1000) )
  {
    a[1] = Tx10 % 10 | PSET;                         // x 0,1 + разделительная точка (0х80)
    a[2] = (Tx10/10) % 10;                           // единицы 
    a[3] = (Tx10/100) % 10;                          // десятки
  }

  if ( Tx10 >= 1000 )
  {
    a[1] = (Tx10 /10) % 10;                         // единицы 
    a[2] = (Tx10 /100) % 10;                        // десятки
    a[3] = (Tx10 /1000) % 10;                       // сотни
  }

  if ( (Tx10 < 0) && (Tx10 > -100) )                
  {
    a[1] = (-Tx10 ) % 10 | PSET;                     // x 0,1 + разделительная точка (0х80)
    a[2] = (-Tx10 /10) % 10;                         // единицы
    a[3] = '-';                                      // -
  }
  
  if ( Tx10 <= -100 )
  {
    a[0] = -Tx10 % 10 | PSET;                         // x 0,1 + разделительная точка (0х80)
    a[1] = (-Tx10/10) % 10;                           // единицы
    a[2] = (-Tx10/100) % 10;                          // десятки
    a[3] = '-';                                       // -
  } 
    
    Write_One_Digit (a[0], 0);
    Write_One_Digit (a[1], 1);
    Write_One_Digit (a[2], 2);
    Write_One_Digit (a[3], 3);
} 


// *******************************************************************************************************************************
// РАБОТА С СЕМИСЕГМЕНТНЫМ ИНДИКАТОРОМ
// *******************************************************************************************************************************

// инициализация семисегментного индикатора
void ind_setup()
{
  byte a[8] = {0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88};
  byte b[8]  = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '};
  ind_displey(a, 8, 1);        // проверка светимости всех сегментов индикатора
  delay(2000);
  ind_displey(b, 8, 1);        
}

// отображение данных на семисегментном индикаторе
void ind_displey(byte *ind_array, byte count, byte min_on_dig)
{
  pinMode(IND_D, OUTPUT);
  pinMode(IND_C1, OUTPUT);
  pinMode(IND_C2, OUTPUT);
  digitalWrite(IND_D, LOW);
  digitalWrite(IND_C1, LOW);
  //digitalWrite(IND_C2, LOW);    // зажигаем индикатор на 100% яркости, остановка ШИМ
  digitalWrite(IND_C2, HIGH);     // гасим индикатор, остановка ШИМ
  
  if (min_on_dig < 1) min_on_dig = 1;                   // минимум 1 разряд всегда светится
    
  for (byte i=1; i<=count-min_on_dig; i++)              // гашение незначащего нуля 
  {
//    if ( (ind_array[count-i] & 0x7F) > 0x0F) continue;      // если в разряде не шестнадцатиричная цифра, а буква, то пропускаем его обработку (точка не учитывается)
    if ( (ind_array[count-i] & 0x7F) > 9) continue;       // если в разряде не десятичная цифра, а буква, то пропускаем его обработку (точка не учитывается)
//    if (ind_array[count-i] == 0) ind_array[count-i] = ' ';    // если в разряде 0 без точки - затираем его пробелом (гасим)
    else break;
  }
  
  digitalWrite(IND_C2, HIGH); // останавливаем ШИМ - гасим
  
  for (byte k=0; k<count; k++)
  {
    byte temp = *(ind_array+(count-1-k)) ;
    byte s = ind_code[ temp & 0x7F ];
    
    if (temp == '-') s = 0x10; //CODE_MINUS;    // код мунуса
    if (temp == '*') s = 0x17; //CODE_DEGREE;   // код радуса
    if (temp == ' ') s = 0x00; //CODE_SPACE;    // код пробела
    if (temp == 'H') s = 0xB5; //CODE_H;        // код H
    if (temp == 'P') s = 0x37; //CODE_P;        // код P
    if (temp == '.') s = 0x08; //CODE_PSET;     // код .
    
    if (temp & 0x80) s |= 0x08; //CODE_PSET;    // точка 

    for (byte i=0; i<8; i++)                    // побитная выдача в последовательном коде
    {
      digitalWrite(IND_D, LOW);
      if (s & 0x80) digitalWrite(IND_D, HIGH);
    
      digitalWrite(IND_C1, HIGH);
      s <<= 1;  
      digitalWrite(IND_C1, LOW); 
    }   
  } 

  analogWrite(IND_C2, 1023-10*setting.seg_ind_bright);          // запуск ШИМ (% яркости)
}

// вывод времени на семисегментный индикатор
void ind_time(bool dvoetochie)
{
  byte a[4];
    
  a[0] = now_time.minutes % 10;
    if (dvoetochie) a[0] |= 0x80;          // c включенными разделительными точками
  a[1] = (now_time.minutes /10) % 10;
  a[2] = now_time.hours % 10;
  a[3] = (now_time.hours /10) % 10;
    if (a[3] == 0) a[3] = ' ';            // гашение незнач. нуля

  ind_displey(a, 4, 4); 
}

// вывод температуры на семисегментный индикатор
void ind_termo(int Tx10)
{
  byte a[4];
  const byte PSET = 0x80;

  if ((Tx10 > 1250) || (Tx10 < -550))               // если выход за диапазон измерений
  {
    a[0] = '-';
    a[1] = '-';
    a[2] = '-';
    a[3] = '-';
    ind_displey(a, 4, 4); 
    return;
  }

  a[0] = '*';
  
  if ( (Tx10 >= 0) && (Tx10 < 1000) )
  {
    a[1] = Tx10 % 10;                                // x 0,1
    a[2] = (Tx10/10) % 10 | PSET;                    // единицы + разделительная тока (0х80)
    a[3] = (Tx10/100) % 10;                          // десятки
    if (a[3] == 0) a[3] = ' ';                       // гашение незначащего нуля
  }

  if ( Tx10 >= 1000 )
  {
    a[1] = (Tx10 /10) % 10;                         // единицы + разделительная тока (0х80)
    a[2] = (Tx10 /100) % 10;                        // десятки
    a[3] = (Tx10 /1000) % 10;                       // сотни
/*    
    a[0] = (Tx10) % 10;                             // x 0,1
    a[1] = (Tx10 /10) % 10 | 0x80;                  // единицы + разделительная тока (0х80)
    a[2] = (Tx10 /100) % 10;                        // десятки
    a[3] = (Tx10 /1000) % 10;                       // сотни
*/  
  }

  if ( (Tx10 < 0) && (Tx10 > -100) )                
  {
    a[1] = (-Tx10 ) % 10;                            // x 0,1
    a[2] = (-Tx10 /10) % 10 | PSET;                  // единицы
    a[3] = '-';                                      // -
  }
  
  if ( Tx10 <= -100 )
  {
    a[0] = -Tx10 % 10;                                // x 0,1
    a[1] = (-Tx10/10) % 10 | PSET;                    // единицы
    a[2] = (-Tx10/100) % 10;                          // десятки
    a[3] = '-';                                       // -
  }

  ind_displey(a, 4, 4); 
}


// *******************************************************************************************************************************
// ВЕБ-сервер
// *******************************************************************************************************************************

// Заруск сервера
void web_server_start() 
{
  Serial.println("AP-mode start!");
  WiFi.mode(WIFI_AP);             // переключаем в режим точки доступа
  WiFi.softAP(ssid, password);    // стандартный пароль

  Serial.println("Web server start!");
  web_server.on ( "/", HTTP_handleRoot );     // настраиваем веб-сервер
  web_server.onNotFound ( handle_NotFound );
  web_server.begin();                         // запускаем веб-сервер
}

// Оработчик главной страницы сервер
void HTTP_handleRoot(void) 
{
  Serial.println("Web request received!");
  for (int i=0; i<web_server.args(); i++) 
  {
    String aName=web_server.argName(i);
    String aVal=web_server.arg(i);

    if (aVal.length() == 0) continue;     // если пустое поле - не обновляем, он останется какой был

    if (aName=="But2") ESP.reset();
    
    else if (aName=="SS")  {   aVal.toCharArray(setting.ap_ssid, 20); }
    else if (aName=="PSW") {   aVal.toCharArray(setting.ap_pass, 20); }
    else if (aName=="NTP") {   aVal.toCharArray(setting.ntp, 30); }
    else if (aName=="TZ")  {   setting.time_zone = aVal.toInt(); }
    else if (aName=="SP")  {   setting.sync_time_period = aVal.toInt(); }
    else if (aName=="7B")  {   setting.seg_ind_bright = aVal.toInt(); }
    else if (aName=="MB")  {   setting.matrix_ind_bright = aVal.toInt(); }
    else if (aName=="IP")  {   setting.ind_period = aVal.toInt(); }
    else if (aName=="TID") {   setting.temper_int_duration = aVal.toInt(); }
    
    else if (aName=="Y")  {   now_time.year = aVal.toInt(); }
    else if (aName=="Mo") {   now_time.month = aVal.toInt(); }
    else if (aName=="D")  {   now_time.date = aVal.toInt();  }
    else if (aName=="H")  {   now_time.hours = aVal.toInt();  }
    else if (aName=="Mi") {   now_time.minutes = aVal.toInt();  }
    else if (aName=="S")  {   now_time.second = aVal.toInt();   }   
  }
  if ( web_server.args() )
  {
    // подменяем нтп-время значением из веб-интерфейса устанавливаем время из веб
    ntp_time = 946684800UL;               // секунд до 2000 года
    ntp_time += now_time.second; 
    ntp_time += 60UL * now_time.minutes;
    ntp_time += 3600UL * now_time.hours; 
    ntp_time += 86400UL * (now_time.date-1); 
    ntp_time += 31536000UL * now_time.year;  // без учета дней високосных лет
    ntp_time += ((now_time.year-1) / 4 + 1 ) * 86400UL;   // высчитываем дополнительные дни високосных годов до текущего
    
    if (now_time.month == 2) ntp_time += 2678400UL; // добавляем дни января
    else if (now_time.month == 3) ntp_time += 5097600UL; // добавляем дни января, февраля (28д)
    else if (now_time.month == 4) ntp_time += 7776000UL; // добавляем дни января, февраля (28д), марта
    else if (now_time.month == 5) ntp_time += 10368000UL;
    else if (now_time.month == 6) ntp_time += 13046400UL;
    else if (now_time.month == 7) ntp_time += 15638400UL;
    else if (now_time.month == 8) ntp_time += 18316800UL;
    else if (now_time.month == 9) ntp_time += 20995200UL;
    else if (now_time.month == 10) ntp_time += 23587200UL;
    else if (now_time.month == 11) ntp_time += 26265600UL;
    else if (now_time.month == 12) ntp_time += 28857600UL;  // добавляем дни января..ноября
    
    if ( ((now_time.year % 4)==0) && (now_time.month > 2) )  ntp_time += 86400UL;  // если текущий год кратен 4 (високосный) и месяц больше февраля, то еще день прибавляем
      
    for (byte i=0; i<MAX7219_COUNT; i++) my_MAX7219.setIntensity(i, setting.matrix_ind_bright);   //  устанавливаем интенсивность свечения от  0 до 16
    
    autonom_time = now_time;  
    ds3231_set_of_NTP();      // установливаем время в микросхеме дс3231
    ms_syncro = millis();     //запоминаем момент установки по внутренней шкале (нужно для вычисления автономного времени)
 
    EEPROM.put(0, setting);     // сохраняем настройки в память
    EEPROM.commit();            // сохраняем изменения в памяти
    Serial.println("Web request processed!");
  }

String out =
"<!DOCTYPE html><html><meta charset=\"utf-8\" /> <form method=get name=ptemp>\
<style> body { background-color: #88d3fa; font-family: Arial, Helvetica, Sans-Serif; Color: #000000; } </style><table align=""center"">\
<tr><td>SSID</td><td> <input maxlength=""20"" name=""SS"" size=""20"" type=""text"" value="+(String)setting.ap_ssid+"></td><td>(20 max)</td></tr>\
<tr><td>Pass</td><td> <input maxlength=""20"" name=""PSW"" size=""20"" type=""text""></td><td>(20 max)</td></tr>\
<tr><td>NTP</td><td> <input maxlength=""30"" name=""NTP"" size=""20"" type=""text"" value="+(String)setting.ntp+"></td><td>(30 max)</td></tr>\
<tr><td>TimeZ</td><td> <input maxlength=""2"" name=""TZ"" size=""10"" type=""text"" value="+(String)setting.time_zone+"></td><td>(0..23)</td></tr>\
<tr><td>SynPer</td><td> <input maxlength=""3"" name=""SP"" size=""10"" type=""text"" value="+(String)setting.sync_time_period+"></td><td>(1..255)</td></tr>\
<tr><td>SegBri</td><td> <input maxlength=""3"" name=""7B"" size=""10"" type=""text"" value="+(String)setting.seg_ind_bright+"></td><td>(1..100)</td></tr>\
<tr><td>MatBri</td><td> <input maxlength=""2"" name=""MB"" size=""10"" type=""text"" value="+(String)setting.matrix_ind_bright+"></td><td>(0..15)</td></tr>\
<tr><td>IndPer</td><td> <input maxlength=""3"" name=""IP"" size=""10"" type=""text"" value="+(String)setting.ind_period+"></td><td>(1..255)</td></tr>\
<tr><td>TDur</td><td> <input maxlength=""3"" name=""TID"" size=""10"" type=""text"" value="+(String)setting.temper_int_duration+"></td><td>(1..255)</td></tr>\
<tr><td>Year</td><td> <input maxlength=""2"" name=""Y"" size=""10"" type=""text"" value="+(String)now_time.year+"></td><td> (0..99)</td></tr>\
<tr><td>Month</td><td> <input maxlength=""2"" name=""Mo"" size=""10"" type=""text"" value="+(String)now_time.month+"></td><td>(1..12)</td></tr>\
<tr><td>Date</td><td> <input maxlength=""2"" name=""D"" size=""10"" type=""text"" value="+(String)now_time.date+"></td><td>(1..31)</td></tr>\
<tr><td>Hour</td><td> <input maxlength=""2"" name=""H"" size=""10"" type=""text""></td><td>(0..23)</td></tr>\
<tr><td>Min</td><td> <input maxlength=""2"" name=""Mi"" size=""10"" type=""text""></td><td>(0..59)</td></tr>\
<tr><td>Sec</td><td> <input maxlength=""2"" name=""S"" size=""10"" type=""text""></td><td>(0..59)</td></tr>\
<tr><td colspan=""2"" align=""center""><input name=""But1"" type=""submit"" value=""SET""></td>\
<td align=""center""><input name=""But2"" type=""submit"" value=""REBOOT""></td></tr>\</tr></table></html>";

  web_server.send ( 200, "text/html", out );
}


void handle_NotFound()
{
  web_server.send(404, "text/plain", "Not found");
  Serial.println("Web: 404 Not found!");
}


// *******************************************************************************************************************************
// ГЛАВНЫЕ ФУНКЦИИ
// *******************************************************************************************************************************

void setup()
{
  Serial.begin(115200);
  Serial.println("\nESP8266 started");

  EEPROM.begin(sizeof(setting));
  EEPROM.get(0, setting);
  
  search_ds18b20();
  ind_setup();
  read_T_ds18b20();
     
// инициализация всех подключенных Dot matrix module MAX7219
  for (byte i=0; i<MAX7219_COUNT; i++) 
  { 
    my_MAX7219.shutdown(i, false);   // 
    my_MAX7219.setIntensity(i, setting.matrix_ind_bright);   //  устанавливаем интенсивность свечения от  0 до 16
    my_MAX7219.clearDisplay(i);     //  очистка дисплея от мусора
  }
 
   Wire.pins(DS3231_SDA, DS3231_SCL);         // пины подключения микросхемы часов
   Wire.setClock(400000);                     // Устанавливает тактовую частоту обмена данными по I2C интерфейсу.
   Wire.begin();

   ds3231_read(true);
   calc_time();
   print_time();
   ind_time(true);

// Соединение с WiFi   
   if( ConnectWiFi(setting.ap_ssid, setting.ap_pass) )
   {
    // Инициализауия UDP соединения с NTP сервером
      udp.begin(localPort);
    
    // синхронизация времени
      while (err_count < 10) // 10 попыток получить время
      {
        err_count++;
        if( GetNTP(setting.ntp) )
        {
          ds3231_set_of_NTP();
      
          calc_time();
          print_time();
          ind_time(true);
       
          // время и температуру в последовательный порт
          Serial.print(" Unix time = "); Serial.print(ntp_time); Serial.print(" => ");
          Serial.print(now_time.date); Serial.print(".");
          Serial.print(now_time.month); Serial.print(".");
          Serial.print("20"); Serial.print(now_time.year);
          
          switch  (now_time.day)
          {
            case 1 : Serial.print(" Пн "); break;
            case 2 : Serial.print(" Вт "); break;
            case 3 : Serial.print(" Ср "); break;
            case 4 : Serial.print(" Чт "); break;
            case 5 : Serial.print(" Пт "); break;
            case 6 : Serial.print(" Сб "); break;
            case 7 : Serial.print(" Вс "); break;
          }
          
          Serial.print(now_time.hours); Serial.print(":");
          Serial.print(now_time.minutes); Serial.print(":");
          Serial.println(now_time.second); 
          break;
        }//if
      }//while
      if (err_count >=10) Serial.println("NTP connect false");    // если сделали 10 неудачных попыток
   }

  web_server_start();
}


void loop()
{
   cur_ms       = millis();
   t_cur        = cur_ms/1000;

   web_server.handleClient();

   if ( ( (t_cur % setting.ind_period)==0) && (ds18b20_count) ) // каждые TEMPER_IND_PERIOD [сек], если кол-во датчикоков не нулевой, отображаем температуры в течение 2 сек.
   {
      read_T_ds18b20();
      if (ds18b20_Tx10[0] >= -550) { ind_termo(ds18b20_Tx10[0]);  print_termo(ds18b20_Tx10[0]); delay(1000*setting.temper_int_duration); }
      if (ds18b20_Tx10[1] >= -550) { ind_termo(ds18b20_Tx10[1]);  print_termo(ds18b20_Tx10[1]); delay(1000*setting.temper_int_duration); }
   }

// Каждые 0.5 секунды выдаем время
   if ( cur_ms < ms1 || (cur_ms - ms1) > 500 )
   {
       ms1 = cur_ms;
       points = !points;

       calc_time();
       print_time();
       ind_time(points);

       if (points)  // каждую секунду выдаем:
       {
          Serial.print("ds_t = "); Serial.print(ds3231_time.hours); Serial.print(":"); Serial.print(ds3231_time.minutes); Serial.print(":"); Serial.print(ds3231_time.second); 
          calc_time();
          Serial.print(" / a_t = "); Serial.print(autonom_time.hours); Serial.print(":"); Serial.print(autonom_time.minutes); Serial.print(":"); Serial.print(autonom_time.second); 
          
          if (ds3231_time.temper >= -55) { Serial.print("   Tin= "); Serial.print(ds3231_time.temper); Serial.print(" *C"); }
          if (ds18b20_Tx10[0] >= -550) { Serial.print(" T1= "); Serial.print((float)ds18b20_Tx10[0]/10); Serial.print(" *C"); }
          if (ds18b20_Tx10[1] >= -550) { Serial.print(" T2= "); Serial.print((float)ds18b20_Tx10[1]/10); Serial.print(" *C"); }
          Serial.println();
       }
   }
  if (setting.sync_time_period == 0) return;
  if ( (t_cur > setting.sync_time_period*3600) ) ESP.reset(); // время больше периода синхронизации и больше нуля - перезагружаемся и заново синхронизируемся
} 
